# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import copy
import time
import numpy as np
import logging
from .meta_pruner import MetaPruner


class Pruner(MetaPruner):

    def __init__(self, model, args, logger, runner):
        super(Pruner, self).__init__(model, args, logger, runner)
        self.logger = logger

    def prune(self):
        self._get_kept_wg_L1()
        self._prune_and_build_new_model()

        if self.args.reinit:
            if self.args.reinit == 'orth':
                self.logger.info('==> Reinit model: orthogonal initialization')
                for module in self.model.modules():
                    if isinstance(module, (nn.Conv2d, nn.Linear)):
                        nn.init.orthogonal_(module.weight.data)

        return self.model
